# -*- coding: utf-8 -*-
"""
Module defining CtrlCraftData class.
"""
from __future__ import absolute_import

from builtins import str
from builtins import object
import os
import psycopg2

import pglite as pl

from .settings import DB_NAME, INITDB_SQL
from .logger import CtrlCraftLogger

if not pl.check_cluster():
    pl.init_cluster()
pl.start_cluster()


class CtrlCraftData(object):
    """
    Interface with Postgres.
    """

    def __init__(self):
        self.__con = None
        self.last_probe_date = None
        self.srid = 32632
        self.conn_string = pl.cluster_params() + \
            " dbname={}".format(DB_NAME)

    def is_valid(self):
        """
        Checks if the underlying database is ready and well formed.
        """

        sql = ('SELECT EXISTS ('
               'SELECT 1 FROM information_schema.tables '
               'WHERE  table_schema = \'ctrlcraft\'  '
               'AND    table_name = \'panels\')')

        valid = False
        if self.__con and self.__fetch_one(sql):
            valid = True

        return valid

    def create(self,values=['1','2','3','4']):
        """
        Creates the database.
        """

        pl.create_db(DB_NAME)
        if self.open():
            self.__init_db()
            for i in values:
                self.truck_create('{}'.format(i))

    def clear(self):
        """
        Clears the database.
        """

        self.close()
        self.__remove_db()
        self.create()

    def open(self):
        """
        Opens the database.
        """

        # Try to connect
        try:
            self.__con = psycopg2.connect(self.conn_string)
        except Exception as e:
            CtrlCraftLogger().log('Open database failed: {}'.format(str(e)))
            self.__con = None
            return False

        CtrlCraftLogger().log('Open database succedded')
        return True

    def close(self):
        """
        Closes the database.
        """

        self.__con.close()
        self.__con = None

    def export(self, filename):
        """
        Exports the database to a dump file
        """

        pl.export_db('ctrlcraft', filename)

    def import_dump(self, filename):
        """
        Imports the dump file.
        """

        if self.is_valid():
            self.close()
            self.__remove_db()

        pl.import_db(filename, 'ctrlcraft')

    def translate(self, customProj,inOut=True):
        tables = ['"co3"','estimated_blocs','probe','slabs']
        if inOut == False :
            customProj = [-i for i in customProj]
        for table in tables :
            sql = ("""UPDATE ctrlcraft.{} SET geom=(SELECT ST_Translate(geom, {},{},{}));""".format(table,customProj[0],customProj[1],customProj[2]))
            self.__commit(sql)
    #
    # Gate methods
    #
    def gate_create(self, deposit, panel, blast, truck, datetime, radiometry):
        """
        Create a new entry for the gate.
        """

        sql = ('insert into ctrlcraft.gate(deposit, panel, blast, truck, time,'
               ' radiometry) values(\'{}\', \'{}\', \'{}\', \'{}\', '
               '\'{}\', {})'
               .format(deposit, panel, blast, truck, datetime, radiometry))
        self.__commit(sql)

    def gate_data(self, truck, datetime):
        """
        Returns dataid if a data already exists with these parameters, false
        otherwise.
        """

        sql = ('select * from ctrlcraft.gate where '
               ' truck = \'{}\' and time = \'{}\''
               .format(truck, datetime))

        cur = self.__con.cursor()
        cur.execute(sql)

        return cur.fetchone()

    def gate_data_entry(self, deposit, panel, blast, truck, unload_time):
        """
        Returns a specific gate entry.
        """

        sql =('select id, radiometry from ctrlcraft.gate where deposit = \'{}\''
              ' and panel = \'{}\' and blast = \'{}\' and truck = \'{}\''
              ' and time >= \'{}\' order by time ASC limit 1'
              .format(deposit, panel, blast, truck, unload_time))

        cur = self.__con.cursor()
        cur.execute(sql)

        return cur.fetchone()

    #
    # Shipments (slabs loaded in trucks) methods
    #
    def shipments(self, deposit, panel, blast):
        """
        Returns all slabs loaded in trucks within deposit/panel/blast.
        """

        sql = ('select count, deposit, panel, blast, truck, unload_time, '
               'radiometry, radiometry_probe, slabs '
               'from ctrlcraft.truck_loads_status '
               'where panel = \'{panel}\' '
               'and blast = \'{blast}\''
               .format(panel=panel, blast=blast))

        cur = self.__con.cursor()
        cur.execute(sql)

        return cur.fetchall()

    def get_max_slabs_shipment(self, deposit, panel, blast):
        """
        Returns all slabs loaded in trucks within deposit/panel/blast.
        """

        sql = ('select count '
               'from ctrlcraft.truck_loads_status '
               'where panel = \'{panel}\' '
               'and blast = \'{blast}\''
               .format(panel=panel, blast=blast))

        cur = self.__con.cursor()
        cur.execute(sql)
        res = [int(i[0])for i in cur.fetchall()]
        if len(res)>0:
            res = max(res)
        else :
            res = 0
        return res

    #
    # Slabs methods
    #
    def slab_create(self, smuid, deposit, panel, blast, quality, lith, perc, unit, x, y, z):
        """
        Creates a slabs thanks to a SMU centroid.
        """

        sql = ('select ctrlcraft.insert_smu(\'{}\', \'{}\', \'{}\', \'{}\', '
               '{}, \'{}\', {}, \'{}\', {}, {}, {})'
               .format(smuid, deposit, panel, blast, quality, lith,
                       perc, unit, x, y, z))
        self.__commit(sql)

    def slabs_create_index(self):
        sql = ('CREATE INDEX IF NOT EXISTS smus_geom_idx ON '
               'ctrlcraft.smus USING gist (geom)')
        self.__commit(sql)

        sql = ('CREATE INDEX IF NOT EXISTS slabs_geom_idx ON '
               'ctrlcraft.slabs USING gist (geom)')
        self.__commit(sql)

    def estimated_blocs_stats_for_slabs(self, deposit, panel, blast):
        """
        Returns estimated blocs statistics for all slabs within deposit/panel/blast.
        """

        sql = ('select ps.radiometry::text as rad_gamma, p.radiometry::text '
               'as rad_panel, ps.sample_number::text as id from '
               'ctrlcraft.estimated_blocs_stats ps join '
               'ctrlcraft.estimated_blocs p on '
               'ps.sample_number = p.sample_number where panel=\'{panel}\' '
               'and blast = \'{blast}\' and deposit=\'{deposit}\' '
               .format(panel=panel, blast=blast, deposit=deposit))

        cur = self.__con.cursor()
        cur.execute(sql)
        records = cur.fetchall()

        return records

    def slab_wkt(self, slab_id):
        """
        Returns the position in WKT for the corresponding slab.
        """

        sql = ('select st_astext(st_envelope(geom)) from ctrlcraft.slabs '
               'where id = \'{slab_id}\''
               .format(slab_id=slab_id))

        return self.__fetch_all(sql)[0]

    def slab_radiometry(self, x, y):
        """
        Returns the radiometry of the upper slab at the position (x, y).
        """

        sql = ('select radiometry, radiometry_probe from '
               'ctrlcraft.slabs_available '
               'where st_intersects(geom, st_setsrid(st_point({}, {}), {})) '
               'order by centroid_z DESC limit 1'
               .format(x, y, self.srid))

        cur = self.__con.cursor()
        cur.execute(sql)

        return cur.fetchone()

    def smu(self, smu_id, deposit, panel, blast):
        """
        Returns the smu id if found in database.
        """

        sql = ('select id from ctrlcraft.smus '
               'where smu_id = \'{smu_id}\' '
               'and panel=\'{panel}\' '
               'and blast=\'{blast}\''
               .format(smu_id=smu_id, blast=blast,
                       deposit=deposit, panel=panel))

        cur = self.__con.cursor()
        cur.execute(sql)

        return cur.fetchone()

    #
    # Estimated Blocs methods
    #
    def estimated_bloc(self, sample_number):
        """
        Returns an estimated bloc from a sample_number.
        """

        sql_string = ('select sample_number from'
                      ' ctrlcraft.estimated_blocs where sample_number '
                      '= \'{sample_number}\''
                      .format(sample_number=sample_number))

        cur = self.__con.cursor()
        cur.execute(sql_string)

        return cur.fetchone()

    def estimated_bloc_wkt(self, sample_number):
        """
        Returns the WKT position for a specific bloc.
        """

        sql_string = ('select st_astext(st_envelope(geom)) from '
                      'ctrlcraft.estimated_blocs where sample_number '
                      '= \'{sample_number}\''
                      .format(sample_number=sample_number))

        cur = self.__con.cursor()
        cur.execute(sql_string)
        records = cur.fetchall()

        return records[0][0]

    def estimated_bloc_create(self, sample_number, radiometry, x, y, z):
        """
        Creates a new estimated bloc.
        """

        sql = ('select ctrlcraft.insert_estimated_bloc({}, {}, {}, {}, {})'
               .format(sample_number, radiometry, x, y, z))
        self.__commit(sql)

    def estimated_bloc_create_index(self):
        """
        Creates a new index for estimated blocs geometries.
        """

        sql = ('CREATE INDEX IF NOT EXISTS estimated_blocs_geom_idx on'
               ' ctrlcraft.estimated_blocs using gist (geom)')
        self.__commit(sql)

    def estimated_bloc_stats(self):
        """
        Computes statistics on panels and blasts for estimated blocs.
        """

        sql = ('DELETE FROM ctrlcraft.estimated_blocs_stats')
        self.__commit(sql)

        sql = ('select ctrlcraft.estimated_blocs_stats_compute()')
        self.__commit(sql)

    #
    # Lithology methods
    #
    def lithologies(self):
        """
        Returns a list of available lithologies.
        """

        sql = 'select unnest(enum_range(NULL::lithology))'

        records = []
        for lithology in self.__fetch_all(sql):
            records.append(lithology)

        return records

    def lithology_update(self, name, point):
        """
        Updates the lithology of the more above slab at the given position.
        """

        wkt = point.asWkt()
        name = name
        sql = ('select ctrlcraft.lithology_update(\'{}\', \'{}\')'
               .format(name, wkt))
        self.__commit(sql)

    def affect_radiometry_slab(self, isUpSonde):
        """
        Affect the radiometry of a slabe using last positions from the GeoCube.
        """
        sql = ("SELECT ctrlcraft.add_fix_measure({}, {})".format(int(isUpSonde),
                                                             int(not
                                                                 isUpSonde)))
        self.__commit(sql)

    def affect_radiometry_slab_degraded(self, teneur_up, teneur_low, wkt):
        """
        Affect the radiometry of a slabe using last positions from the GeoCube.
        Degraded mode.
        """
        sql = ("SELECT ctrlcraft.add_fix_measure_degraded({}, {}, '{}')".format(teneur_up, teneur_low, wkt))

        self.__commit(sql)

    def get_average_last_four_radiometry(self):
        """
        Returns the average of last 4 radioemtries recorded
        """
        sql = ("""WITH radio as
                  (SELECT teneurup, teneurdown FROM ctrlcraft.probe
                  ORDER BY "time"
                  LIMIT 4)
                  SELECT avg(teneurup) as up, avg(teneurdown) as down
                  FROM radio""")

        return self.__fetch_one(sql, entireRow=True)
    #
    # Deposit methods
    #
    def deposits(self):
        """
        Returns a list of all deposits.
        """

        sql = 'select id from ctrlcraft.deposits order by id'

        return self.__fetch_all(sql)

    #
    # Panel methods
    #
    def panels(self, depositid):
        """
        Returns a list of all panels in the deposit.
        """

        sql = ('select id from ctrlcraft.panels '
               'where deposit = \'{}\' order by id'
               .format(depositid))

        return self.__fetch_all(sql)

    #
    # Blast methods
    #
    def blasts(self, panelid):
        """
        Returns a list of all blasts.
        """

        sql = ('select id from ctrlcraft.blasts where panel = \'{}\' order '
               'by id'
               .format(panelid))

        return self.__fetch_all(sql)

    #
    # Trucks methods
    #
    def truck_create(self, name):
        """
        Creates a new truck.
        """

        sql = ('insert into ctrlcraft.trucks values(\'{}\')'.format(name))
        self.__commit(sql)

    def trucks(self):
        """
        Returns a list of trucks.
        """

        sql = 'select id from ctrlcraft.trucks order by id'

        return self.__fetch_all(sql)

    def trucks_no_order(self):
        """
        Returns a list of trucks.
        """

        sql = 'select id from ctrlcraft.trucks'

        return self.__fetch_all(sql)

    def truck_slabs(self, truckid):
        """
        Returns slabs currently loaded in the truck.
        """
        if truckid in self.trucks():
            sql = ('select array_length(loads, 1) from ctrlcraft.trucks '
                   'where id = \'{}\''
                   .format(truckid))

            rec = self.__fetch_all(sql)[0]
            if not rec:
                rec = 0
        else :
            rec = 0

        return rec

    def truck_load_slab(self, truckid, point):
        """
        Loads a slab in the specific truck.
        """

        wkt = point.asWkt()
        sql = ('select ctrlcraft.truck_load_slab(\'{}\', \'{}\')'
               .format(truckid, wkt))
        self.__commit(sql)

    def truck_unload_slabs(self, truckid):
        """
        Unloads all slabs in the specific truck.
        """

        sql = ('select ctrlcraft.truck_unload_slabs(\'{}\')'
               .format(truckid))

        self.__commit(sql)

    def truck_unload_last_slab(self):
        """
        Unload the last slab loaded in a truck.
        """

        sql = 'select ctrlcraft.truck_unload_last_slab()'

        return self.__fetch_one(sql, True)

    def truck_ids_change(self,new_ids):
        """
        Change trucks ID when the parameters is changed
        """
        current_ids = self.trucks()
        for ci in current_ids:
            sql= "delete from ctrlcraft.trucks where id='{}'".format(ci)
            self.__commit(sql)
        for ni in new_ids:
            self.truck_create('{}'.format(ni))
    #
    # Settings methods
    #
    def settings_probe_min_date_update(self):
        """
        Updates the settings date for probe with the current date.
        """

        if self.last_probe_date:
            sql = ('update ctrlcraft.settings set value = \'{}\' '
                   'where name = \'probe_min_date\''
                   .format(self.last_probe_date))
            self.__commit(sql)

    #
    # Bluetooth spare : mode dégradé
    def slab_update_probe_radiometry(self, x, y, radiometry):
        """
        Add radiometry when position is not known
        """
        sql = ('SELECT ctrlcraft.slab_update_probe_radiometry({}, {}, {})'.format(x, y, radiometry))
        self.__commit(sql)

    #
    # Probe methods
    #
    def probe_new_data(self, date, lat, lon, alt, radiometry, gammaup, gammadown,
                       teneurup, teneurdown):
        """
        Returns the most recent positions of the probe.
        """

        self.last_probe_date = date
        sql = ('select ctrlcraft.probe_new_data(\'{}\', {}, {}, {}, {}, {}, {}, {}, {})'
               .format(date, float(lat), float(lon), float(alt), float(radiometry),
                       float(gammaup), float(gammadown), float(teneurup),
                       float(teneurdown)))
        self.__commit(sql)

    def probe_export(self, filename):
        """
        Exports probe data in a CSV file.
        """

        sql = ('select ctrlcraft.probe_to_csv(\'{}\')'
               .format(filename))
        self.__execute(sql)

    #
    # CO3 methods
    #
    def co3(self, x, y):
        sql = ('select * from ctrlcraft.co3'
               ' where st_x(geom) = {x} and st_y(geom) = {y} limit 1'
               .format(x=x, y=y))

        cur = self.__con.cursor()
        cur.execute(sql)

        return cur.fetchone()

    def co3_create_index(self):
        """
        Creates a gist index on co3 data.
        """

        sql = ('CREATE INDEX IF NOT EXISTS co3_geom_idx ON '
               'ctrlcraft.co3 USING gist (geom)')
        self.__commit(sql)

    def co3_create(self, deposit, panel, blast, depth, co3, x, y, z):
        """
        Creates a new CO3 hole.
        """

        geom = ('St_SetSrid(ST_MakePoint({x},{y},{z}), {srid})'
                .format(x=x, y=y, z=z, srid=self.srid))

        sql = ('INSERT INTO ctrlcraft.co3 VALUES(\'{}\', \'{}\', \'{}\', {}, '
               '{}, {})'
               .format(deposit, panel, blast, depth, co3, geom))
        self.__execute(sql)

    #
    # Internal methods
    #
    def __execute(self, sql):
        cur = self.__con.cursor()
        cur.execute(sql)

    def __fetch_one(self, sql, commit=False, entireRow=False):
        cur = self.__con.cursor()
        cur.execute(sql)
        if entireRow:
            res = cur.fetchone()
        else:
            res = cur.fetchone()[0]

        if commit:
            self.__con.commit()

        return res

    def __fetch_all(self, sql):
        cur = self.__con.cursor()
        cur.execute(sql)

        records = []
        for record in cur.fetchall():
            records.append(record[0])

        return records

    def __commit(self, sql):
        cur = self.__con.cursor()
        cur.execute(sql)
        self.__con.commit()

    def __init_db(self):
        statements = []
        dirname = os.path.dirname(os.path.abspath(__file__))
        initdb_sql = os.path.join(dirname, 'utils', INITDB_SQL)
        with open(initdb_sql) as sqlfile:
            statements = sqlfile.read().split('\n;\n')[:-1]

        for statement in statements:
            self.__commit(statement)

    def __close_sessions(self):
        con_str = 'dbname=postgres {}'.format(pl.cluster_params())
        con = psycopg2.connect(con_str)
        cur = con.cursor()
        con.set_isolation_level(0)
        cur.execute('select pg_terminate_backend(pg_stat_activity.pid)'
                    ' from pg_stat_activity'
                    ' where pg_stat_activity.datname = \'{}\''
                    .format(DB_NAME))

        cur.execute('select count(1) from pg_catalog.pg_database where'
                    ' datname=\'{}\''.format(DB_NAME))
        con.commit()
        con.close()

    def __remove_db(self):
        self.__close_sessions()
        pl.drop_db(DB_NAME)
