# -*- coding: utf-8 -*-
"""
Truck class
"""
from __future__ import absolute_import

import os.path

from PyQt5.QtGui import QPixmap, QFont, QPainter, QIcon, QColor, QBrush
from PyQt5.QtCore import QPointF, QRectF, pyqtSignal, QObject, QSettings

from qgis.gui import QgsMapToolEmitPoint

from .simulator import CtrlCraftSimulator
from .settings import SIMU

settings = QSettings()


class CtrlCraftTruckLoadMapTool(QgsMapToolEmitPoint):
    """
    Map tool allowing to add a slab in a truck.
    """

    def __init__(self, truck):
        QgsMapToolEmitPoint.__init__(self, truck.master.canvas)
        self.truck = truck

    def canvasReleaseEvent(self, event):
        if self.truck.isFull():
            msg = ('This truck is filled, you have to empty it first.')
            self.truck.master.iface.messageBar().pushWarning('Warning', msg)
        else:
            self.truck.addLoad(event.mapPoint())
            self.truck.master.refresh()

            if self.truck.isFull():
                msg = 'The truck {} is filled.'.format(self.truck.id)
                self.truck.master.iface.messageBar().pushInfo('Infos', msg)


class CtrlCraftTruck(QObject):
    """
    Truck class
    """

    updated = pyqtSignal()

    def __init__(self, id, master):
        """
        Constructor.
        """
        QObject.__init__(self)
        self.id = id
        self.max_slab = self.getMaxSlab()
        self.master = master
        self.__actionLoad = None
        self.__actionUnload = None
        self.mapToolLoad = None

        self.init()

    @staticmethod
    def unloadLastSlab(master):
        if master.data.is_valid():
            id = master.data.truck_unload_last_slab()
            master.trucks[id].updateIconLoad()
            master.trucks[id].updateIconUnload()
            master.refresh()

    @property
    def loads(self):
        return self.master.data.truck_slabs(self.id)

    @property
    def actionLoad(self):
        return self.__actionLoad

    @actionLoad.setter
    def actionLoad(self, action):
        """ Associate action fill to truck """
        self.__actionLoad = action
        self.updateIconLoad()

    @property
    def actionUnload(self):
        return self.__actionUnload

    @actionUnload.setter
    def actionUnload(self, action):
        self.__actionUnload = action
        self.updateIconUnload()

    def activateMapTool(self):
        self.mapToolLoad = CtrlCraftTruckLoadMapTool(self)
        self.mapToolLoad.setAction(self.__actionLoad)
        self.master.canvas.setMapTool(self.mapToolLoad)

    def removeLoads(self):
        self.master.data.truck_unload_slabs(self.id)
        self.updateIconLoad()
        self.updateIconUnload()

        if SIMU:
            deposit = self.master.currentDeposit
            panel = self.master.currentPanel
            blast = self.master.currentBlast

            CtrlCraftSimulator().gate_create(deposit, panel, blast, self.id)

    def addLoad(self, point):
        """ Add a load in the truck """
        if self.loads < self.max_slab:
            # Mode dégradé / Bluetooth spare
            if all([x == '-1' for x in [self.master.rtkube.widget.GNSSLat.text(), self.master.rtkube.widget.GNSSLon.text(), self.master.rtkube.widget.GNSSAlt.text()]]):
                self.master.data.slab_update_probe_radiometry(point.x(), point.y(), int(self.master.rtkube.widget.gammaGUp.text()) / float(settings.value('/GeoCube/gammaup', 1000.00)))
            self.master.data.truck_load_slab(self.id, point)
            self.master.data.settings_probe_min_date_update()
            self.updateIconLoad()
            self.updateIconUnload()

        return self.isFull()

    def updateIconLoad(self):
        """ Update fill button icon """
        if not self.actionLoad:
            return

        name = self.loadIcon()
        path = os.path.join(os.path.dirname(__file__), 'img', name)
        pix = QPixmap(path)
        font = QFont()
        font.setPointSizeF(10)
        textPix = self.id

        position = QPointF(0, 60)
        painter = QPainter()
        painter.begin(pix)
        painter.setFont(font)
        painter.drawText(position, textPix)

        width = pix.width()
        if not self.isFull():
            width = (pix.width()/self.max_slab) * self.loads
        rect = QRectF(0.0, 0.0, width, pix.height())
        brush = QBrush(QColor(255, 0, 0, 150))
        painter.fillRect(rect, brush)

        painter.end()

        icon = QIcon(pix)

        self.__actionLoad.setIcon(icon)

        self.updated.emit()

    def updateIconUnload(self):
        """
        Update empty button icon
        """

        if not self.actionUnload:
            return

        iconName = self.unloadIcon()

        pix = QPixmap(os.path.join(os.path.dirname(__file__), 'img', iconName))
        font = QFont()
        font.setPointSizeF(10)
        textPix = self.id

        position = QPointF(0, 60)
        painter = QPainter()
        painter.begin(pix)
        painter.setFont(font)
        painter.drawText(position, textPix)
        painter.end()

        icon = QIcon(pix)

        self.actionUnload.setIcon(icon)

        self.actionUnload.setEnabled(not self.isEmpty())

    def loadIcon(self):
        """ Returns the initial fill icon """
        return "icon_truck.png"

    def unloadIcon(self):
        """ Returns the empty icon """
        return "icon_truck_empty.png"

    def init(self):
        """ Init the state of the truck """
        self.updateIconLoad()
        self.updateIconUnload()

    def getMaxSlab(self):
        """ Retrieve slab capacity from params.ini file"""
        max_slab = 8
        for ind in range(1,5):
            if 'Truck {}'.format(int(settings.value('CtrlCraft/Truck{}/id'.format(ind), ind))) == str(self.id):
                max_slab = int(settings.value('CtrlCraft/Truck{}/sc'.format(ind), 8))
        return max_slab

    def isFull(self):
        """ Check if the truck is full """
        return self.loads == self.max_slab

    def isEmpty(self):
        return self.loads == 0
